"""
Login to pathfinder.

>>> from pathfinder.operators.login import Login
>>> c = {"host": "pathfinder", "user": "Admin", "pass": "Admin"}
>>> l = Login(client=client, config=c)
>>> l.execute()
True
>>> c = {"host": "pathfinder", "user": "Admin", "pass": "Nope"}
>>> l = Login(client=client, config=c)
>>> l.execute()
False
"""

from pathfinder.operators.base import BaseOperator


class LoginException(Exception):
    pass


class Login(BaseOperator):
    config = {"user": "Admin", "pass": "Admin"}

    def execute(self) -> bool:
        self.client.write(
            "LOGIN {} {}".format(
                self.config.get("user"), self.config.get("pass")
            ).encode("ascii")
            + b"\r\n"
        )
        index, *_ = self.client.expect(
            [b"login failed", b"login successful"], timeout=1
        )
        return index == 1



from abc import ABC, abstractmethod

from ..telnet import TelnetClient


class BaseOperator(ABC):
    config = {}

    def __init__(self, client: TelnetClient, config={}):
        self.client = client.client
        self.config = {**self.config, **config}

    @abstractmethod
    def execute(self):  # pragma: no cover
        pass



"""Get data from pathfinder."""

from pathfinder.operators.base import BaseOperator
from pathfinder.parser import ResponseParser


class GetException(Exception):
    pass


class Get(BaseOperator):
    config = {"path": "."}

    def execute(self):
        p = ResponseParser()
        self.client.write(
            "GET {}".format(self.config.get("path")).encode("ascii") + b"\r\n"
        )
        results = []
        while True:
            data = self.client.read_until(b"\r\n", timeout=1)
            if data == b"":
                break
            if data != b"\r\n":
                results.append(p.parse(data.decode("ascii").rstrip("\r\n")))
        return results



"""
Wrapper around telnetlib.

>>> from pathfinder.telnet import TelnetClient
>>> tc = TelnetClient()
>>> tc.connect()
>>> tc.disconnect()
"""

import logging
import telnetlib

logger = logging.getLogger(__name__)


class TelnetClientException(Exception):
    """Exceptions raised by TelnetClient."""

    pass


class TelnetClient:
    """Telnet client handler."""

    config = {"host": "localhost", "port": 9600, "timeout": 1}

    def __init__(self, config={}):
        self.config = {**self.config, **config}
        self.client = telnetlib.Telnet()

    def connect(self):
        """Connect the client."""
        try:
            self.client.open(
                self.config.get("host"),
                self.config.get("port"),
                self.config.get("timeout"),
            )
        except Exception as ex:
            logger.error("Error connecting telnet client")
            raise TelnetClientException(ex)

    def disconnect(self):
        """Disconnect the client."""
        try:
            self.client.close()
        except Exception as ex:
            logger.error("Error disconnecting telnet client")
            raise TelnetClientException(ex)



"""
SUB operator

https://docs.telosalliance.com/pathfinder-core-pro/pathfindercore-pro/appendix-a-sapv2#subscription-examples
"""

from typing import NoReturn

from pathfinder.operators.base import BaseOperator


class Sub(BaseOperator):
    config = {"path": "."}

    def execute(self) -> NoReturn:
        self.client.write(f"SUB {self.config.get('path')}".encode("ascii") + b"\r\n")



from parsy import regex, seq, string, string_from


class ResponseParser:
    def __init__(self):
        # TODO: finish this parser to it does more than just parse indi and GPO reponses
        indi_operator = string_from("indi")
        gp_operator = string_from("GPO", "GPI")
        space = string(" ")
        obj = regex(r"[a-zA-Z0-9.#\[\]:/]*")
        name = regex(r"[a-zA-Z]*")
        simple_string = regex(r"[a-zA-Z ]*")
        equals = string("=")
        value = regex(r"[^,]*") | regex(r'".*"')
        number = regex(r"[0-9]+")
        gp_value = regex(r"[hl]").map(lambda v: {"h": False, "l": True}[v]) * 5

        indi_parser = seq(indi_operator << space).then(
            seq(
                path=obj << space.optional(),
                info=seq(name=name << equals, value=value << string(", ").optional())
                .map(lambda x: {x["name"]: x["value"]})
                .many()
                .map(lambda kv: {k: v for d in kv for k, v in d.items()}),
            )
        )
        gp_parser = seq(gp_operator << space).then(
            seq(number=number << space, pins=gp_value)
        )
        error_parser = seq(string("ERROR") << space).then(
            seq(number << space, simple_string)
        )

        self.p = indi_parser | gp_parser | error_parser

    def parse(self, string):
        return self.p.parse(string)
